@@ -22,9 +22,11 @@
  * \24, \25, \26 - an unsigned byte immediate operand, from operand 0, 1 or 2
  * \30, \31, \32 - a word immediate operand, from operand 0, 1 or 2
  * \34, \35, \36 - select between \3[012] and \4[012] depending on 16/32 bit
- *                 assembly mode or the address-size override on the operand
+ *                 assembly mode or the operand-size override on the operand
  * \37           - a word constant, from the _segment_ part of operand 0
  * \40, \41, \42 - a long immediate operand, from operand 0, 1 or 2
+ * \44, \45, \46 - select between \3[012] and \4[012] depending on 16/32 bit
+ *                 assembly mode or the address-size override on the operand
  * \50, \51, \52 - a byte relative operand, from operand 0, 1 or 2
  * \60, \61, \62 - a word relative operand, from operand 0, 1 or 2
  * \64, \65, \66 - select between \6[012] and \7[012] depending on 16/32 bit
@@ -72,6 +74,7 @@
 #include "nasmlib.h"
 #include "assemble.h"
 #include "insns.h"
+#include "preproc.h"
 
 extern struct itemplate *nasm_instructions[];
 
@@ -87,8 +90,8 @@ static efunc errfunc;
 static struct ofmt *outfmt;
 static ListGen *list;
 
-static long calcsize (long, long, int, insn *, char *);
-static void gencode (long, long, int, insn *, char *, long);
+static long calcsize (long, long, int, insn *, const char *);
+static void gencode (long, long, int, insn *, const char *, long);
 static int  regval (operand *o);
 static int  matches (struct itemplate *, insn *);
 static ea * process_ea (operand *, ea *, int, int, int);
@@ -99,7 +102,7 @@ static int  chsize (operand *, int);
  * in order to pass a copy of the data off to the listing file
  * generator at the same time.
  */
-static void out (long offset, long segto, void *data, unsigned long type,
+static void out (long offset, long segto, const void *data, unsigned long type,
 		 long segment, long wrt) 
 {
     long lineno;
@@ -150,7 +153,7 @@ static void out (long offset, long segto, void *data, unsigned long type,
 }
 
 static int jmp_match (long segment, long offset, int bits,
-		insn *ins, char *code)
+		insn *ins, const char *code)
 {   long isize;
     unsigned char c = code[0];
 
@@ -265,6 +268,8 @@ long assemble (long segment, long offset, int bits, unsigned long cp,
 	static char fname[FILENAME_MAX];
 	FILE        * fp;
 	long        len;
+        char *prefix = "", *combine;
+        char** pPrevPath = NULL;
 
 	len = FILENAME_MAX-1;
 	if (len > instruction->eops->stringlen)
@@ -272,7 +277,26 @@ long assemble (long segment, long offset, int bits, unsigned long cp,
 	strncpy (fname, instruction->eops->stringval, len);
 	fname[len] = '\0';
 
-	if ( (fp = fopen(fname, "rb")) == NULL)
+        while (1)             /* added by alexfru: 'incbin' uses include paths */
+        {
+          combine = nasm_malloc(strlen(prefix) + len + 1);
+          strcpy(combine, prefix);
+          strcat(combine, fname);
+
+          if ( (fp = fopen(combine, "rb")) != NULL)
+          {
+              nasm_free(combine);
+              break;
+          }
+
+          nasm_free(combine);
+          pPrevPath = pp_get_include_path_ptr (pPrevPath);
+          if (pPrevPath == NULL)
+              break;
+          prefix = *pPrevPath;
+        }
+
+        if (fp == NULL)
 	    error (ERR_NONFATAL, "`incbin': unable to open file `%s'", fname);
 	else if (fseek(fp, 0L, SEEK_END) < 0)
 	    error (ERR_NONFATAL, "`incbin': unable to seek on file `%s'",
@@ -347,14 +371,13 @@ long assemble (long segment, long offset, int bits, unsigned long cp,
 
 	if (m == 100) 		       /* matches! */
 	{
-	    char *codes = temp->code;
+	    const char *codes = temp->code;
 	    long insn_size = calcsize(segment, offset, bits,
 				      instruction, codes);
 	    itimes = instruction->times;
 	    if (insn_size < 0)	       /* shouldn't be, on pass two */
 	    	error (ERR_PANIC, "errors made it through from pass one");
 	    else while (itimes--) {
-		insn_end = offset + insn_size;
 		for (j=0; j<instruction->nprefix; j++) {
 		    unsigned char c=0;
 		    switch (instruction->prefixes[j]) {
@@ -370,6 +393,10 @@ long assemble (long segment, long offset, int bits, unsigned long cp,
 		      case R_FS: c = 0x64; break;
 		      case R_GS: c = 0x65; break;
 		      case R_SS: c = 0x36; break;
+		      case R_SEGR6:
+		      case R_SEGR7:
+			error (ERR_NONFATAL, "segr6 and segr7 cannot be used as prefixes");
+			break;
 		      case P_A16:
 			if (bits != 16)
 			    c = 0x67;
@@ -396,6 +423,7 @@ long assemble (long segment, long offset, int bits, unsigned long cp,
 			offset++;
 		    }
 		}
+		insn_end = offset + insn_size;
 		gencode (segment, offset, bits, instruction, codes, insn_end);
 		offset += insn_size;
 		if (itimes > 0 && itimes == instruction->times-1) {
@@ -483,13 +511,35 @@ long insn_size (long segment, long offset, int bits, unsigned long cp,
 	char  fname[FILENAME_MAX];
 	FILE  * fp;
 	long  len;
+        char *prefix = "", *combine;
+        char** pPrevPath = NULL;
 
 	len = FILENAME_MAX-1;
 	if (len > instruction->eops->stringlen)
 	    len = instruction->eops->stringlen;
 	strncpy (fname, instruction->eops->stringval, len);
 	fname[len] = '\0';
-	if ( (fp = fopen(fname, "rb")) == NULL )
+
+        while (1)             /* added by alexfru: 'incbin' uses include paths */
+        {
+          combine = nasm_malloc(strlen(prefix) + len + 1);
+          strcpy(combine, prefix);
+          strcat(combine, fname);
+
+          if ( (fp = fopen(combine, "rb")) != NULL)
+          {
+              nasm_free(combine);
+              break;
+          }
+
+          nasm_free(combine);
+          pPrevPath = pp_get_include_path_ptr (pPrevPath);
+          if (pPrevPath == NULL)
+              break;
+          prefix = *pPrevPath;
+        }
+
+        if (fp == NULL)
 	    error (ERR_NONFATAL, "`incbin': unable to open file `%s'", fname);
 	else if (fseek(fp, 0L, SEEK_END) < 0)
 	    error (ERR_NONFATAL, "`incbin': unable to seek on file `%s'",
@@ -521,7 +571,7 @@ long insn_size (long segment, long offset, int bits, unsigned long cp,
 	if (m == 100) {
 	    /* we've matched an instruction. */
 	    long  isize;
-	    char  * codes = temp->code;
+	    const char  * codes = temp->code;
 	    int   j;
 
 	    isize = calcsize(segment, offset, bits, instruction, codes);
@@ -564,7 +614,7 @@ static int is_sbyte (insn *ins, int op, int size)
 }
 
 static long calcsize (long segment, long offset, int bits,
-		      insn *ins, char *codes) 
+		      insn *ins, const char *codes) 
 {
     long           length = 0;
     unsigned char  c;
@@ -590,12 +640,18 @@ static long calcsize (long segment, long offset, int bits,
       case 030: case 031: case 032:
 	length += 2; break;
       case 034: case 035: case 036:
-	length += ((ins->oprs[c-034].addr_size ?
-		    ins->oprs[c-034].addr_size : bits) == 16 ? 2 : 4); break;
+	if ( ins->oprs[c-034].type & (BITS16|BITS32) )
+	  length += (ins->oprs[c-034].type & BITS16) ? 2 : 4;
+	else
+	  length += (bits == 16) ? 2 : 4;
+	break;
       case 037:
 	length += 2; break;
       case 040: case 041: case 042:
 	length += 4; break;
+      case 044: case 045: case 046:
+	length += ((ins->oprs[c-044].addr_size ?
+		    ins->oprs[c-044].addr_size : bits) == 16 ? 2 : 4); break;
       case 050: case 051: case 052:
 	length++; break;
       case 060: case 061: case 062:
@@ -670,7 +726,7 @@ static long calcsize (long segment, long offset, int bits,
 }
 
 static void gencode (long segment, long offset, int bits,
-		     insn *ins, char *codes, long insn_end) 
+		     insn *ins, const char *codes, long insn_end) 
 {
     static char condval[] = { /* conditional opcodes */
 	0x7, 0x3, 0x2, 0x6, 0x2, 0x4, 0xF, 0xD, 0xC, 0xE, 0x6, 0x2,
@@ -799,9 +855,11 @@ static void gencode (long segment, long offset, int bits,
 	    break;
 
 	case 034: case 035: case 036:
+	    if ( ins->oprs[c-034].type & (BITS16|BITS32) )
+	      size = (ins->oprs[c-034].type & BITS16) ? 2 : 4;
+	    else
+	      size = (bits == 16) ? 2 : 4;
 	    data = ins->oprs[c-034].offset;
-	    size = ((ins->oprs[c-034].addr_size ?
-		     ins->oprs[c-034].addr_size : bits) == 16 ? 2 : 4);
 	    if (size==2 && (data < -65536L || data > 65535L))
 		errfunc (ERR_WARNING, "word value exceeds bounds");
 	    out (offset, segment, &data, OUT_ADDRESS+size,
@@ -827,6 +885,17 @@ static void gencode (long segment, long offset, int bits,
 	    offset += 4;
 	    break;
 
+	case 044: case 045: case 046:
+	    data = ins->oprs[c-044].offset;
+	    size = ((ins->oprs[c-044].addr_size ?
+		     ins->oprs[c-044].addr_size : bits) == 16 ? 2 : 4);
+	    if (size==2 && (data < -65536L || data > 65535L))
+		errfunc (ERR_WARNING, "word value exceeds bounds");
+	    out (offset, segment, &data, OUT_ADDRESS+size,
+		 ins->oprs[c-044].segment, ins->oprs[c-044].wrt);
+	    offset += size;
+	    break;
+
 	case 050: case 051: case 052:
 	    if (ins->oprs[c-050].segment != segment)
 		errfunc (ERR_NONFATAL, "short relative jump outside segment");
@@ -885,7 +954,8 @@ static void gencode (long segment, long offset, int bits,
 	case 0130: case 0131: case 0132:
 	    data = ins->oprs[c-0130].offset;
 	    if (is_sbyte(ins, c-0130, 16)) {
-		out (offset, segment, &data, OUT_RAWDATA+1, NO_SEG, NO_SEG);
+	        bytes[0] = data;
+		out (offset, segment, bytes, OUT_RAWDATA+1, NO_SEG, NO_SEG);
 		offset++;
 	    } else {
 		if (ins->oprs[c-0130].segment == NO_SEG &&
@@ -910,7 +980,8 @@ static void gencode (long segment, long offset, int bits,
 	case 0140: case 0141: case 0142:
 	    data = ins->oprs[c-0140].offset;
 	    if (is_sbyte(ins, c-0140, 32)) {
-		out (offset, segment, &data, OUT_RAWDATA+1, NO_SEG, NO_SEG);
+	        bytes[0] = data;
+		out (offset, segment, bytes, OUT_RAWDATA+1, NO_SEG, NO_SEG);
 		offset++;
 	    } else {
 		out (offset, segment, &data, OUT_ADDRESS+4,
@@ -1082,37 +1153,14 @@ static void gencode (long segment, long offset, int bits,
 	}
 }
 
+#include "regvals.c"
+
 static int regval (operand *o) 
 {
-    switch (o->basereg) {
-      case R_EAX: case R_AX: case R_AL: case R_ES: case R_CR0: case R_DR0:
-      case R_ST0: case R_MM0: case R_XMM0:
-	return 0;
-      case R_ECX: case R_CX: case R_CL: case R_CS: case R_DR1: case R_ST1:
-      case R_MM1: case R_XMM1:
-	return 1;
-      case R_EDX: case R_DX: case R_DL: case R_SS: case R_CR2: case R_DR2:
-      case R_ST2: case R_MM2:  case R_XMM2:
-	return 2;
-      case R_EBX: case R_BX: case R_BL: case R_DS: case R_CR3: case R_DR3:
-      case R_TR3: case R_ST3: case R_MM3:  case R_XMM3:
-	return 3;
-      case R_ESP: case R_SP: case R_AH: case R_FS: case R_CR4: case R_TR4:
-      case R_ST4: case R_MM4:  case R_XMM4:
-	return 4;
-      case R_EBP: case R_BP: case R_CH: case R_GS: case R_TR5: case R_ST5:
-      case R_MM5:  case R_XMM5:
-	return 5;
-      case R_ESI: case R_SI: case R_DH: case R_DR6: case R_TR6: case R_ST6:
-      case R_MM6:  case R_XMM6:
-	return 6;
-      case R_EDI: case R_DI: case R_BH: case R_DR7: case R_TR7: case R_ST7:
-      case R_MM7:  case R_XMM7:
-	return 7;
-      default:			       /* panic */
-        errfunc (ERR_PANIC, "invalid register operand given to regval()");
-	return 0;
+    if ( o->basereg < EXPR_REG_START || o->basereg >= REG_ENUM_LIMIT ) {
+	errfunc (ERR_PANIC, "invalid operand passed to regval()");
     }
+    return regvals[o->basereg];
 }
 
 static int matches (struct itemplate *itemp, insn *instruction) 
@@ -1289,6 +1337,9 @@ static ea *process_ea (operand *input, ea *output, int addrbits, int rfield,
 		if (((s==2 && i!=R_ESP && !(input->eaflags & EAF_TIMESTWO)) ||
 		     s==3 || s==5 || s==9) && b==-1)
 		    b = i, s--;       /* convert 3*EAX to EAX+2*EAX */
+		if (i==-1 && b!=R_ESP && (input->eaflags & EAF_TIMESTWO))
+		    i = b, b = -1, s = 1;
+			/* convert [NOSPLIT EAX] to sib format with 0x0 displacement */
 		if (s==1 && i==R_ESP)  /* swap ESP into base if scale is 1 */
 		    i = b, b = R_ESP;
 		if (i==R_ESP || (s!=1 && s!=2 && s!=4 && s!=8 && i!=-1))
