@@ -32,7 +32,7 @@ static int get_bits (char *value);
 static unsigned long get_cpu (char *cpu_str);
 static void parse_cmdline (int, char **);
 static void assemble_file (char *);
-static int getkw (char *buf, char **value);
+static int getkw (char **directive, char **value);
 static void register_output_formats(void);
 static void report_error_gnu (int severity, const char *fmt, ...);
 static void report_error_vc (int severity, const char *fmt, ...);
@@ -378,6 +378,8 @@ static int process_arg (char *p, char *q)
 	  case 'E':
 	  case 'F':
 	  case 'X':
+	  case 'u':
+	  case 'U':
 	    if ( !(param = get_param (p, q, &advance)) )
 		break;
 	    if (p[1]=='o') {	       /* output file */
@@ -401,13 +403,15 @@ static int process_arg (char *p, char *q)
 			while(isdigit(*++param)) ;
 			if (opt<=0) optimizing = -1;		/* 0.98 behaviour */
 			else if (opt==1) optimizing = 0;	/* Two passes, 0.98.09 behavior */
-			else if (opt<=3) optimizing = opt*5;	/* Multiple passes */
 			else optimizing = opt;			/* Multiple passes */
 		    } else {
 			if (*param == 'v' || *param == '+') {
 			    ++param;
 			    opt_verbose_info = TRUE;
 			    opt = 0;
+			} else {	/* garbage */
+			    opt = -99;
+			    break;
 			}
 		    }
 		} /* while (*param) */
@@ -777,7 +781,7 @@ static void parse_cmdline(int argc, char **argv)
 
 static void assemble_file (char *fname)
 {
-    char   * value, * p, * q, * special, * line, debugid[80];
+    char   * directive, * value, * p, * q, * special, * line, debugid[80];
     insn   output_ins;
     int    i, rn_error, validid;
     long   seg, offs;
@@ -829,7 +833,8 @@ static void assemble_file (char *fname)
 
          /* here we parse our directives; this is not handled by the 'real'
             * parser. */
-         if ( (i = getkw (line, &value)) )
+         directive = line;
+         if ( (i = getkw (&directive, &value)) )
          {
                switch (i) {
                case 1:               /* [SEGMENT n] */
@@ -1054,10 +1059,10 @@ static void assemble_file (char *fname)
                       }
                   break;
                default:
-                  if (!ofmt->directive (line+1, value, pass2))
+                  if (!ofmt->directive (directive, value, pass2))
                      report_error (pass1==1 ? ERR_NONFATAL : ERR_PANIC,
                               "unrecognised directive [%s]",
-                              line+1);
+                              directive);
                }
          }
          else         /* it isn't a directive */
@@ -1304,9 +1309,11 @@ static void assemble_file (char *fname)
 } /* exit from assemble_file (...) */
 
 
-static int getkw (char *buf, char **value)
+static int getkw (char **directive, char **value)
 {
-    char *p, *q;
+    char *p, *q, *buf;
+    
+    buf = *directive;
 
     /*  allow leading spaces or tabs */
     while (*buf==' ' || *buf=='\t')
@@ -1331,7 +1338,7 @@ static int getkw (char *buf, char **value)
     }
     q[1] = '\0';
 
-    p = buf+1;
+    *directive = p = buf+1;
     while (*buf && *buf!=' ' && *buf!=']' && *buf!='\t')
     	buf++;
     if (*buf==']') {
